/*
 * GestioninterventionsSessionBehavior.java
 * Copyright (c) 2011, W4.
 * All rights reserved.
 */

package GestionInterventions.behavior;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;

import leon.app.LyDataSession;
import leon.app.LySession;
import leon.control.LyController;
import leon.data.LyObject;
import leon.info.LyAction;
import leon.misc.LyEnvironment;
import leon.view.web.LyWebSession;
import w4.demo.common.STRINGS;
import w4.demo.common.behavior.*;

/**
 * This interface defines the methods to implement in order to change the
 * default behavior of a session.
 *
 * @author W4 S.A.
 * @see leon.app.LySession
 */
public class GestioninterventionsSessionBehavior extends SessionBehavior
{

  /**
   * Default Constructor.
   *
   * @see eu.w4.leon.appmodel.appdefault.behavior.W4SessionBehavior#W4SessionBehavior
   */
  public GestioninterventionsSessionBehavior() {
    super();
  }  

  /**
   * Gets user of this session.
   *
   * @param session
   *          Current session.
   *
   * @return The object representing current user of the session.
   */
  public LyObject getUser(LyDataSession session)
  {
    final LyObject user = super.getUser(session);
    if (user != null)
    {
      return user;
    }
    session.addValue(STRINGS.USER_DOMAIN, getStoreUser(session));

    final String login = (String)session.getValue(STRINGS.USER_LOGIN_AUTO);

    if (login == null || "".equals(login))
    {
      return user;
    }

    return getUserFromLogin(session, login);
  }
  

  /*
   * Retrieves the authenticated user (W4 Store) by using the
   * cookie named "DemoUser"
   */
  private String getStoreUser(LyDataSession session)
  {
    LyEnvironment env = session.getApplication().getEnvironment();
    if (env.getEnv("FORCE_DOMAIN")!=null && !env.getEnv("FORCE_DOMAIN").equals(""))
    {
      env.logWarning("[getStoreUser] FORCE_DOMAIN with "+env.getEnv("FORCE_DOMAIN"));
      return env.getEnv("FORCE_DOMAIN");
    }
    else {
    env.logTrace("[getStoreUser] appel mthode");
    String storeUsr = null;
    if (session instanceof LyWebSession)
    {
      LyWebSession ws = (LyWebSession) session;
      HttpServletRequest req = ws.getCurrentRequest();
      if (req != null)
      {
        Cookie[] cookies = req.getCookies();
        if (cookies !=null)
        {
          for (int i = 0; i < cookies.length; i++)
        {
          if (cookies[i].getName() != null && cookies[i].getName().equals(STRINGS.DEMO_USER))
          {
            storeUsr = cookies[i].getValue();
            break;
          }
        }
        if (req.getAttribute(STRINGS.DEMO_USER) != null)
          storeUsr = req.getAttribute(STRINGS.DEMO_USER).toString();
        }
      }
    }
    env.logTrace("Login Drupal : " + storeUsr);
    return storeUsr;
    }
  }




  @Override
  public boolean validateAction(LySession session, LyAction action, LyController controller)
  {
    LyObject currentUser = session.getUser();
    if (currentUser != null)
    {
      if (((String)session.getValue(STRINGS.USER_DOMAIN)!=null)
          && (action.hasMark(STRINGS.ACCESS_LIMITED)))
      {
        return false;
      }
    }
    return super.validateAction(session, action, controller);
  }
  
}
