/*
 * InterventionClassBehavior.java
 * Copyright (c) 2011, W4.
 * All rights reserved.
 */

package GestionInterventions.behavior;

import leon.app.behavior.LyClassBehavior;
import leon.control.LySetController;
import leon.data.LyObject;
import leon.data.LyObjectId;
import leon.data.LyRelationValue;
import leon.data.LyValueSet;
import leon.info.LyClassInfo;
import leon.info.LyRelationFieldInfo;
import GestionInterventions.src.GESTIONINTERVENTIONS;

/**
 * Generic interface for all behaviors that applies on action. This interface represents the default
 * behavior for consult, set, filter and delete actions.
 *
 * @author  Lyria S.A.
 * @version $Id: LyClassBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.app.behaviorinterface.LySetBehaviorInterface
 * @see     leon.app.behaviorinterface.LyCreateBehaviorInterface
 * @see     leon.app.behaviorinterface.LyDeleteBehaviorInterface
 * @see     leon.app.behaviorinterface.LyFilterBehaviorInterface
 */
public class InterventionClassBehavior extends LyClassBehavior
{
  /**
   * Default Constructor
   */
  public InterventionClassBehavior()
  {
  }

  @Override
  public void addExtraValues(LySetController set, LyValueSet values) {

    super.addExtraValues(set, values);
    if (set.isCreation())
    {

      LyClassInfo interventionCls = set.getClassInfo();
      LyRelationFieldInfo interAddressFld = (LyRelationFieldInfo)interventionCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_INTERVENTION_ADDRESS);
      LyRelationValue customerValue =(LyRelationValue) values.getFieldValue(GESTIONINTERVENTIONS.FLD_INTERVENTION_CLI);
      LyObject customer = (customerValue != null && customerValue.getValueCount() == 1 ? customerValue.getObject(0) : null);
      if(customer != null)
      {
        LyObjectId addressId = (LyObjectId)customer.getValue(GESTIONINTERVENTIONS.FLD_CUSTOMER_ADDRESS);

        if(addressId != null)
        {
          LyObject address = addressId.getLogicalObject();

          if (address != null)
          {
            LyClassInfo addressCls = address.getClassInfo();
            LyObject newAddress = addressCls.createInstance();

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_ADDRESS),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_ADDRESS));

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_CITY),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_CITY));

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_COUNTRY),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_COUNTRY));

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_ZIPCODE),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_ZIPCODE));

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_LATITUDE),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_LATITUDE));

            newAddress.setValue(
                                addressCls.getFieldInfo(GESTIONINTERVENTIONS.FLD_ADDRESS_LONGITUDE),
                                address.getValue(GESTIONINTERVENTIONS.FLD_ADDRESS_LONGITUDE));
            set.getControllerContext().createObject(newAddress);
            values.add(new LyRelationValue(interAddressFld, newAddress.getObjectId()));
          }
        }
      }
    }

  }

}
