package GestionInterventions.command;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import leon.app.LyApplication;
import leon.app.LyProcessActionInterface;
import leon.app.LySession;
import leon.control.LyController;
import leon.data.LyClassWorkSpace;
import leon.data.LyObject;
import leon.data.LyObjectList;
import leon.data.LySimpleObjectList;
import leon.data.LyValueSet;
import leon.info.LyAction;
import leon.info.LyClassInfo;
import leon.info.LyFieldInfo;
import leon.info.LyFilter;
import leon.info.LyFilterExpression;
import leon.info.LyInfoList;
import leon.info.LyObjectInfo;
import leon.info.LyRelationFieldInfo;
import leon.peer.util.LyRowObject;
import eu.w4.ae.peer.w4.W4BpmDataProvider;
import eu.w4.ae.peer.w4.W4BpmRequestEncoder;
import eu.w4.bpm.BPMDataLocator;
import eu.w4.bpm.BPMDataType;
import eu.w4.bpm.BPMException;
import eu.w4.bpm.BPMVariable;
import eu.w4.bpm.BPMVariables;
import eu.w4.bpm.search.BPMDomainFilter;
import eu.w4.bpm.service.BPMBusinessTypeService;
import eu.w4.bpm.service.BPMDomainSnapshot;
import eu.w4.bpm.service.BPMService;
import eu.w4.bpm.service.BPMSessionId;
import eu.w4.leon.appmodel.appdefault.W4IncludesIds;
import eu.w4.leon.appmodel.appdefault.tools.W4Tools;


/**
 * Handle the new workcase action.
 */
public class W4NewWorkcaseCommand_forDomain
implements LyProcessActionInterface
{
  public LyController processAction(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
                                    LyObjectList objects)
  {
    LyApplication       app = session.getApplication();
    LyClassInfo         domainCls = W4IncludesIds.clsW4Domain;
    W4BpmRequestEncoder encoder = W4BpmDataProvider.getBpmRequestEncoder(domainCls);

    if (app.isEdited())
    {
      LyAction                  createWorkcaseOnDomainsChoiceAction = 
        W4IncludesIds.actW4NewWorkcaseInDomain;
      LyAction                  creationOnDomainAction;
      Hashtable<String, Object> params = new Hashtable<String, Object>(1);

      params.put(W4IncludesIds.PROCESS_ID_PARAM, "*");
      params.put(W4IncludesIds.WORKCASE_ID_PARAM, "*");
      creationOnDomainAction = new LyAction(createWorkcaseOnDomainsChoiceAction, params);
      return session.invokeAction(parent, creationOnDomainAction, domainCls, null, false);
    }

    String      processId;
    LyClassInfo   processClass = null;
    LyObjectInfo  template = action.getTemplate();

    boolean startsFromBusinessData = (template != null
        && (template == W4IncludesIds.actW4NewWorkcaseFromBusinessData));

    if(startsFromBusinessData)
    {
      processClass = (LyClassInfo)action.getParameter(W4IncludesIds.PROCESS_CLASS_PARAM);
      processId = session.getEnvironment().getString(processClass.getName());
      if(processId.contains(" rev."))
        processId = processId.substring(0, processId.indexOf(" rev."));
    }
    else
    {
      processId = (String)action.getParameter(W4IncludesIds.PROCESS_ID_PARAM, true);

      if (processId == null)
        processId = (String)action.getParameter(W4IncludesIds.WORKCASE_ID_PARAM, true);
    }

    Vector<LyRowObject> possibleDomains = encoder.getUpperDomainsOnWhichUserCanStartWorkcase(processId);

    //ajout code spcifique pour gestion automatique du domaine issu de Drupal
    if (((String)session.getValue("$USER_DOMAIN")!=null))
    {
      try 
      {
        final BPMService service = W4Tools.getBPMService(parent.getSession());
        BPMSessionId localsid = service.getSessionService().openSession("w4adm", "bypass".getBytes());
        //final BPMSessionId sessionId = W4Tools.getBPMSessionId(parent.getSession());
        BPMDomainFilter filter = service.getDomainService().createDomainFilter();
        filter.domainNameLike((String)session.getValue("$USER_DOMAIN"));
        List<BPMDomainSnapshot> domainsW4 = service.getDomainService().searchDomains(localsid, filter, null);
        if (domainsW4.isEmpty())
        {
          service.getDomainService().createDomain(localsid, (String)session.getValue("$USER_DOMAIN"));
        }
      }
      catch (Exception e) {
        e.printStackTrace();
      }
    }
    //fin code spcifique

    if ((possibleDomains == null)
        || (possibleDomains.size() == 0))
    {
      // The user can not create a workcase because he hasn't the required role on domain
      String message = app.getEnvironment().getMessage("MSG_NO_ROLE_INITIATOR_NO_DOMAIN");

      parent.showWarning(message);
      return null;
    }

    Vector<BPMVariable> bpmVars = new Vector<BPMVariable>();
    if(startsFromBusinessData)
    {
      String w4VariableName = getW4VariableName(processClass, classInfo);

      if(w4VariableName != null)
      {
        for(int i=0; i<objects.getSize(); i++)
        {
          LyObject object = objects.getObject(i);
          String id = object.getId();

          final BPMService service = W4Tools.getBPMService(parent.getSession());
          final BPMSessionId sessionId = W4Tools.getBPMSessionId(parent.getSession());
          BPMBusinessTypeService businessTypeService;
          BPMDataLocator locator = null;
          try
          {
            businessTypeService = service.getBusinessTypeService();
            locator = businessTypeService.buildBPMDataLocator(sessionId, null, id, null);
          }
          catch (BPMException e)
          {
            locator = null;
          }

          BPMVariable bpmVar = BPMVariables.createVariable(w4VariableName, BPMDataType.DATA, locator);
          bpmVars.add(bpmVar);
        }
      }
    }

    //ajout code spcifique pour gestion automatique du domaine issu de Drupal
    if ((possibleDomains.size() == 1) || 
        ((String)session.getValue("$USER_DOMAIN")!= null))
    {
      // The user has the role initiator on only one domain, so we don't display the list of domains. We create directly
      // the workcase on the domain.
      String   domainLogicalId; 
      if ((String)session.getValue("$USER_DOMAIN")!= null)
      {
        domainLogicalId =(String)session.getValue("$USER_DOMAIN");
      }
      else {
        domainLogicalId = possibleDomains.get(0).getValues().getFieldValue("getLabel").getString();
      }
      //fin code spcifique

      LyObject newTask = null;

      if(startsFromBusinessData)
      {
        if(bpmVars.size() < 1)
        {
          newTask = W4Tools.createWorkcase(parent, processId, domainLogicalId);

          String message = app.getEnvironment().getMessage("MSG_IMPOSSIBLE_TO_INITIATE_FROM_BUSINESS_DATA");
          parent.showWarning(message);
        }
        else
        {
          for(int i=0; i<bpmVars.size(); i++)
            newTask = W4Tools.createWorkcase(parent, processId, domainLogicalId, bpmVars.get(i), true);

          String taskActor = "";
          if(newTask != null)
          {
            Object actorValue = newTask.getValue("w4_task_actor");
            if (actorValue != null)
              taskActor = actorValue.toString();
          }

          if(bpmVars.size() > 1 || taskActor.startsWith("__"))
            newTask = null;
        }
      }
      else
        newTask = W4Tools.createWorkcase(parent, processId, domainLogicalId);

      if (newTask == null)
      {
        String error = parent.getApplication().getError();

        if (error == null)
        {
          if(objects != null && objects.getSize() > 1)
            parent.showWarning(session.getEnvironment()
                               .getMessage("MSG_NEW_WORKCASES_SUCCESS", String.valueOf(objects.getSize())));
          else
            parent.showWarning("MSG_NEW_WORKCASE_SUCCESS");
        }
        else
          parent.showError("MSG_ER_NEW_WORKCASE\n(" + error + ")");

        return null;
      }

      LyController controller = invokeProcessTaskAction(parent, session, newTask);

      return controller;
    }
    else if (possibleDomains.size() > 1)
    {
      // The user can create a workcase on different domains
      // We display a screen where user can choose the wanted domain on which the workcase will
      // be created.

      LyAction           createWorkcaseOnDomainsChoiceAction;
      LyObjectList    list = new LySimpleObjectList();
      LyValueSet      valueSet = null;
      LyFilter      filter = new LyFilter(LyFilter.Type.OR);
      Object        domainId;
      LyFilterExpression  exp;
      LyFieldInfo     domainIdField = W4IncludesIds.fldW4DomainInternalId;

      for (int i = 0; i < possibleDomains.size(); i++)
      {
        valueSet = possibleDomains.get(i).getValues();

        // Domain Id
        domainId = valueSet.getFieldValue(W4IncludesIds.FLD_W4_DOMAIN_GET_ID_PEER_COLUMN).getValue();

        exp = new LyFilterExpression(domainIdField, domainId);
        filter.addElement(exp);
      }

      list = new LyClassWorkSpace(session, domainCls, filter, false);

      Hashtable<String, Object> params = new Hashtable<String, Object>();

      params.put(W4IncludesIds.PROCESS_ID_PARAM, processId);
      params.put(W4IncludesIds.WORKCASE_ID_PARAM, processId);

      createWorkcaseOnDomainsChoiceAction = W4IncludesIds.actW4NewWorkcaseInDomain;
      LyAction creationOnDomainAction = new LyAction(createWorkcaseOnDomainsChoiceAction, params);

      if(startsFromBusinessData)
      {
        // Pas trs propre, reflechir  une autre mthode
        parent.getSession().addValue(W4IncludesIds.W4_BPM_VARIABLES_PARAM, bpmVars);
      }

      return session.invokeAction(parent, creationOnDomainAction, domainCls, list, true);
    }

    return null;

  }

  protected String getW4VariableName(LyClassInfo processClass, LyClassInfo businessDataClass)
  {
    LyInfoList<LyFieldInfo> fields = processClass.getFieldInfos();

    String result = null;
    int index = 0, cpt = 0;
    while(cpt < 2 && index < fields.getSize())
    {
      LyFieldInfo field = fields.getInfo(index);

      if(!(field instanceof LyRelationFieldInfo))
      {
        index++;
        continue;
      }

      LyRelationFieldInfo relationField = (LyRelationFieldInfo)field;
      if(relationField.getClassInfoCount() != 1)
      {
        index++;
        continue;
      }

      LyClassInfo temp = relationField.getClassInfo(0);
      if(businessDataClass.equals(temp))
      {
        result = relationField.getBinding().getBinding();

        if(result.contains("#"))
          result = result.substring(0, result.indexOf("#"));

        cpt++;
      }

      index++;
    }

    if(cpt != 1)
      return null;

    return businessDataClass.getEnvironment().getString(result);
  }

  /**
   * Processes the given task with activity form.
   *
   * @param   parent
   * @param   session
   * @param   newTask
   *
   * @return
   */
  private LyController invokeProcessTaskAction(LyController parent, LySession session, LyObject newTask)
  {
    LyAction           processTaskAction = W4IncludesIds.actW4ProcessTaskCommand;

    // For compatibility reasons, fall back to the process task action
    if (processTaskAction == null)
      processTaskAction = W4IncludesIds.actW4ProcessTask;

    LySimpleObjectList list = new LySimpleObjectList();

    list.addObject(newTask);

    return session.invokeAction(parent, processTaskAction, newTask.getClassInfo(), list, true);
  }
}
